/*
 * SA11x0 + UCB1x00 Touch Screen Driver Version 0.2
 * Put together by Tak-Shing Chan <tchan.rd@idthk.com>
 *
 * 90% derived from the Touch screen driver for Tifon
 * Copyright 1999 Peter Danielsson
 *
 * Codec routines derived from Itsy's Touchscreen driver
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 * Author: Larry Brakmo.
 *
 * Sample filtering derived from Diagman (ucb1x00Touch.c)
 * Copyright (C) 1999 Intel Corp.
 *
 * IOCTL's and jitter elimination derived from ADS Touchscreen driver
 * Copyright (c) 2000 Century Software, Inc.
 *
 * iPAQ emulation derived from Driver for the h3600 Touch Screen
 * Copyright 2000 Compaq Computer Corporation.
 * Author: Charles Flynn.
 *
 * Add Some function to let it more flexible
 * Allen_cheng@xlinux.com
 * chester@linux.org.tw
 *
 *
 * Add Freebird UCB1300 Touch-Panel/Button driver support
 * Eric Peng <ercipeng@coventive.com>
 * Tony Liu < tonyliu@coventive.com>
 *
 * Todo:
 * support other button driver controlled by UCB-1300
 *
 */

/*
 * INSTRUCTIONS
 * To use this driver, simply ``mknod /dev/h3600_ts c 11 0'' and
 * use it as if it were an iPAQ.  TODO: iPAQ-compatible IOCTL's.
 * 
 *
 */

#include <linux/types.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <linux/string.h>
#include "ucb1200_ts.h"

/* From Compaq's Touch Screen Specification version 0.2 (draft) */
typedef struct {
    short pressure;
    short x;
    short y;
    short millisecs;
} TS_EVENT;

#ifdef CONFIG_SA1100_FREEBIRD
// for button driver
typedef struct {
	char keycode;
	char down;
} BT_EVENT;
#endif

static int raw_max_x, raw_max_y, res_x, res_y, raw_min_x, raw_min_y, xyswap;

static int cal_ok, x_rev, y_rev;

static char *dev_id = "ucb1200";

static wait_queue_head_t queue;
static struct timer_list timer;
static int Reg10;
#define PRESSED 0
#define P_DONE 1
#define X_DONE 2
#define Y_DONE 3
#define RELEASED 4
#define BUFSIZE 128
#define XLIMIT 160
#define YLIMIT 160
static int state, head, tail, sample;
static TS_EVENT cur_data, samples[3], buf[BUFSIZE];
static struct fasync_struct *fasync;
static unsigned long in_timehandle = 0;

/* Allen Add */
static void ts_clear(void);
static void print_par(void);

#ifdef CONFIG_SA1100_FREEBIRD
// for button driver
static wait_queue_head_t bt_queue;
static int bt_head, bt_tail;
static BT_EVENT bt_cur_data, bt_buf[BUFSIZE];
static void bt_clear(void);
static int bt_int_mask = IO0_4_INT;
#endif

static int ucb_irq;

static inline void ts_clear_codec_int(void)
{
   int s;

/* Clear interrupts at codec  */
   if ((s = codec_read(CODEC_REG_INT_STATUS) & (TSPX_INT | ADC_INT)))
   {
      codec_write(CODEC_REG_INT_STATUS, 0);
      codec_write(CODEC_REG_INT_STATUS, s);
   }
}

static inline void set_read_x_pos(void)
{
/* See Philips' AN809 for an explanation of the pressure mode switch */
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

#ifdef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPY | ADC_ENA);
#else
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPY | ADC_ENA | ADC_SYNC_ENA);
#endif

   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_GND | TSC_MODE_POSITION | TSC_BIAS_ENA);

   udelay(50);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

#ifndef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
#endif
}

static inline void set_read_y_pos(void)
{
   codec_write(CODEC_REG_TS_CTL, TSPY_POW | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

#ifdef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA);
#else
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA | ADC_SYNC_ENA);
#endif

   codec_write(CODEC_REG_TS_CTL, TSPY_POW | TSMY_GND | TSC_MODE_POSITION | TSC_BIAS_ENA);

   udelay(50);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

#ifndef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
#endif
}

static inline void set_read_pressure(void)
{
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA);

#ifdef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA);
#else
   codec_write(CODEC_REG_ADC_CTL, Reg10 = ADC_INPUT_TSPX | ADC_ENA | ADC_SYNC_ENA);
#endif

   udelay(50);

   codec_write(CODEC_REG_ADC_CTL, Reg10 | ADC_START);

#ifndef CONFIG_SA1100_CERF
   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) | 0x200);

   codec_write(CODEC_REG_ADC_CTL, Reg10);

   codec_write(CODEC_REG_IO_DATA, codec_read(CODEC_REG_IO_DATA) & ~0x200);
#endif
}

static int read_adc(void)
{
   int v;

   while (!((v = codec_read(CODEC_REG_ADC_DATA)) & ADC_DAT_VAL));
      return GET_ADC_DATA(v);
}

static int sa1100_ts_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
   int ret_data = 0;

/* Microwindows style (should change to TS_CAL when the specification is ready) */
   switch (cmd)
   {
      case 3:
	 raw_max_x = arg;
	 break;
      case 4:
	 raw_max_y = arg;
	 break;
      case 5:
	 res_x = arg;
	 break;
      case 6:
	 res_y = arg;
	 break;
      case 10:
	 raw_min_x = arg;
	 break;
      case 11:
	 raw_min_y = arg;
	 break;
      case 12:
/* New attribute for portrait modes */
	 xyswap = arg;
/* Allen Add */
      case 13:	       /* 0 = Enable calibration ; 1 = Calibration OK */
	 cal_ok = arg;
      case 14:	       /* Clear all buffer data */
	 ts_clear();
	 break;
      case 15:	       /* X axis reversed setting */
	 x_rev = arg;
	 break;
      case 16:	       /* Y axis reversed setting */
	 y_rev = arg;
	 break;
      case 17:	       /* Clear all buffer data */
	 print_par();
	 break;
/* Allen */
   }

   return 0;
}


static void ts_clear(void)
{
   int i;

   for (i=0; i < BUFSIZE; i++)
   {
       buf[i].pressure=(short)NULL;
       buf[i].x=(int)NULL;
       buf[i].y=(int)NULL;
       buf[i].millisecs=(int)NULL;
   }

   head = 0;
   tail = 0;

}

static void print_par(void)
{
   printk(" Kernel ==> cal_ok = %d\n",cal_ok);
   printk(" Kernel ==> raw_max_x = %d\n",raw_max_x);
   printk(" Kernel ==> raw_max_y = %d\n",raw_max_y);
   printk(" Kernel ==> res_x = %d\n",res_x);
   printk(" Kernel ==> res_y = %d\n",res_y);
   printk(" Kernel ==> raw_min_x = %d\n",raw_min_x);
   printk(" Kernel ==> raw_min_y = %d\n",raw_min_y);
   printk(" Kernel ==> xyswap = %d\n",xyswap);
   printk(" Kernel ==> x_rev = %d\n",x_rev);
   printk(" Kernel ==> y_rev = %d\n",y_rev);

   return 0;
}

/* Allen */

static inline int pen_up(void)
{
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND);

   return codec_read(CODEC_REG_TS_CTL) & TSPX_LOW;
}

static void new_data(void)
{
   static TS_EVENT last_data = { 0, 0, 0, 0 };
   int diff0, diff1, diff2, diff3;

   if (cur_data.pressure)
   {
      if (sample < 3)
      {
	 samples[sample].x = cur_data.x;
	 samples[sample++].y = cur_data.y;
	 return;
      }
      sample = 0;

/* Check the variance between X samples (discard if not similar), then choose the closest pair */
      diff0 = abs(samples[0].x - samples[1].x);
      diff1 = abs(samples[1].x - samples[2].x);
      diff2 = abs(samples[2].x - cur_data.x);
      diff3 = abs(cur_data.x - samples[1].x);

      if (diff0 > XLIMIT || diff1 > XLIMIT || diff2 > XLIMIT || diff3 > XLIMIT)
	 return;

      if (diff1 < diff2)
      {
	 if (diff1 < diff3)
	    cur_data.x = (samples[1].x + samples[2].x) / 2;
	 else
	    cur_data.x = (cur_data.x + samples[1].x) / 2;
      }
      else
      {
	 if (diff2 < diff3)
	    cur_data.x = (samples[2].x + cur_data.x) / 2;
	 else
	    cur_data.x = (cur_data.x + samples[1].x) / 2;
      }

/* Do the same for Y */
      diff0 = abs(samples[0].y - samples[1].y);
      diff1 = abs(samples[1].y - samples[2].y);
      diff2 = abs(samples[2].y - cur_data.y);
      diff3 = abs(cur_data.y - samples[1].y);

      if (diff0 > YLIMIT || diff1 > YLIMIT || diff2 > YLIMIT || diff3 > YLIMIT)
	 return;

      if (diff1 < diff2)
      {
	 if (diff1 < diff3)
	    cur_data.y = (samples[1].y + samples[2].y) / 2;
	 else
	    cur_data.y = (cur_data.y + samples[1].y) / 2;
      }
      else
      {
	 if (diff2 < diff3)
	    cur_data.y = (samples[2].y + cur_data.y) / 2;
	 else
	    cur_data.y = (cur_data.y + samples[1].y) / 2;
      }
   }
   else
   {
/* Reset jitter detection on pen release */
      last_data.x = 0;
      last_data.y = 0;
   }

/* Jitter elimination */
/* Allen Mask */
//   if ((last_data.x || last_data.y) && abs(last_data.x - cur_data.x) <= 3 && abs(last_data.y - cur_data.y) <= 3)
//	return;
/* Allen */

   cur_data.millisecs = jiffies;
   last_data = cur_data;

   if (head != tail)
   {
      int last = head--;
      if (last < 0)
	 last = BUFSIZE - 1;
   }
   buf[head] = cur_data;

   if (++head == BUFSIZE)
      head = 0;
   if (head == tail && tail++ == BUFSIZE)
      tail = 0;

   if (fasync)
      kill_fasync(&fasync, SIGIO, POLL_IN);
   wake_up_interruptible(&queue);
}

static TS_EVENT get_data(void)
{
   int last = tail;

   if (++tail == BUFSIZE)
      tail = 0;
   return buf[last];
}

static void wait_for_action(void)
{
   state = PRESSED;
   sample = 0;
#ifdef CONFIG_SA1100_FREEBIRD
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0 + bt_int_mask);
   codec_write(CODEC_REG_FALL_INT_ENABLE, TSPX_INT + bt_int_mask);
#else
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, TSPX_INT);
#endif
   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW | TSPY_GND | TSMY_GND | TSC_MODE_INT);
}

static void start_chain(void)
{
   state = P_DONE;
#ifdef CONFIG_SA1100_FREEBIRD
   codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT + bt_int_mask);
#else
   codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
#endif
   set_read_pressure();
}

static unsigned int sa1100_ts_poll(struct file *filp, struct poll_table_struct *wait)
{
   poll_wait(filp, &queue, wait);
   if (head != tail)
      return POLLIN | POLLRDNORM;
   return 0;
}

static ssize_t sa1100_ts_read(struct file *filp, char *buf, size_t count, loff_t *l)
{
   DECLARE_WAITQUEUE(wait, current);
   int i;
   TS_EVENT t;
   short out_buf[4];

   if (head == tail)
   {
      if (filp->f_flags & O_NONBLOCK)
	 return -EAGAIN;
      add_wait_queue(&queue, &wait);
      current->state = TASK_INTERRUPTIBLE;
      while ((head == tail) && !signal_pending(current))
      {
	 schedule();
	 current->state = TASK_INTERRUPTIBLE;
      }
      current->state = TASK_RUNNING;
      remove_wait_queue(&queue, &wait);
   }
   for (i = count; i >= sizeof(out_buf); i -= sizeof(out_buf), buf += sizeof(out_buf))
   {
      if (head == tail)
	 break;
      t = get_data();

      out_buf[0] = t.pressure;

/* Alen Add */
#if 0

#ifdef CONFIG_SA1100_ASSABET
      if (xyswap)
      {
	 out_buf[1] = (((raw_max_y - t.y)) * res_y) / (raw_max_y - raw_min_y);
	 out_buf[2] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
      }
      else
      {
	 out_buf[1] = (((raw_max_x - t.x)) * res_x) / (raw_max_x - raw_min_x);
	 out_buf[2] = (((raw_max_y - t.y)) * res_y) / (raw_max_y - raw_min_y);
      }
#else
      if (xyswap)
      {
	 out_buf[1] = (((t.y - raw_min_y)) * res_y) / (raw_max_y - raw_min_y);
	 out_buf[2] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
      }
      else
      {
	 out_buf[1] = (((t.x - raw_min_x)) * res_x) / (raw_max_x - raw_min_x);
	 out_buf[2] = (((t.y - raw_min_y)) * res_y) / (raw_max_y - raw_min_y);
      }
#endif

#else
      if (cal_ok)
      {
	 	out_buf[1] = (x_rev) ? ((raw_max_x - t.x) * res_x) / (raw_max_x - raw_min_x) :
		((t.x - raw_min_x) * res_x) / (raw_max_x - raw_min_x);
	 	out_buf[2] = (y_rev) ? ((raw_max_y - t.y) * res_y) / (raw_max_y - raw_min_y) :
		((t.y - raw_min_y) * res_y) / (raw_max_y - raw_min_y);
      }
      else
      {
	 	out_buf[1] = t.x;
	 	out_buf[2] = t.y;
      }
#endif
/* Allen */

      out_buf[3] = t.millisecs;

      copy_to_user(buf, &out_buf, sizeof(out_buf));
   }

   return count - i;
}

/* Forward declaration */
static void sa1100_ts_timer(unsigned long);

static int sa1100_ts_starttimer(void)
{
   in_timehandle++;
   init_timer(&timer);
   timer.function = sa1100_ts_timer;
   timer.expires = jiffies + HZ / 100;
   add_timer(&timer);
   return 0;
}

static void sa1100_ts_timer(unsigned long data)
{
   in_timehandle--;
   if (pen_up())
   {
      cur_data.pressure = 0;
      new_data();
      wait_for_action();
   }
   else
      start_chain();
}

static int sa1100_ts_fasync(int fd, struct file *filp, int on)
{
   int retval;

   retval = fasync_helper(fd, filp, on, &fasync);
   if (retval < 0)
      return retval;
   else
      return 0;
}

static int sa1100_ts_open(struct inode *inode, struct file *filp)
{
/* Allen Add */
   ts_clear();
/* Allen */

   MOD_INC_USE_COUNT;
   return 0;
}

static int sa1100_ts_release(struct inode *inode, struct file *filp)
{
/* Allen Add */
   ts_clear();
/* Allen */

   sa1100_ts_fasync(-1, filp, 0);
   MOD_DEC_USE_COUNT;
   return 0;
}

static void sa1100_ts_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
   ts_clear_codec_int();
   if (in_timehandle > 0)
      return;
   switch (state)
   {
      case PRESSED:
	 start_chain();
	 break;
      case P_DONE:
	 cur_data.pressure = read_adc();
#ifdef CONFIG_SA1100_FREEBIRD
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT + bt_int_mask);
	 codec_write(CODEC_REG_FALL_INT_ENABLE, 0 + bt_int_mask);
#else
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
	 codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
#endif
	 set_read_x_pos();
	 state++;
	 break;
      case X_DONE:
	 cur_data.x = read_adc();
#ifdef CONFIG_SA1100_FREEBIRD
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT + bt_int_mask);
#else
	 codec_write(CODEC_REG_RISE_INT_ENABLE, ADC_INT);
#endif
	 set_read_y_pos();
	 state++;
	 break;
      case Y_DONE:
	 cur_data.y = read_adc();
#ifdef CONFIG_SA1100_FREEBIRD
	 codec_write(CODEC_REG_RISE_INT_ENABLE, TSPX_INT + bt_int_mask);
#else
	 codec_write(CODEC_REG_RISE_INT_ENABLE, TSPX_INT);
#endif
	 codec_write(CODEC_REG_ADC_CTL, 0);
	 codec_write(CODEC_REG_TS_CTL, TSPX_POW | TSMX_POW |
	 TSPY_GND | TSMY_GND | TSC_MODE_INT);
	 state++;
	 new_data();
	 sa1100_ts_starttimer();
	 break;
      case RELEASED:
	 cur_data.pressure = 0;
	 new_data();
	 wait_for_action();
   }
}

static struct file_operations sa1100_ts_fops = {
   read: sa1100_ts_read,
   poll: sa1100_ts_poll,
   ioctl: sa1100_ts_ioctl,
   fasync: sa1100_ts_fasync,
   open: sa1100_ts_open,
   release: sa1100_ts_release,
};

int sa1100_ts_init(void)
{
//   unsigned int irq;

#ifdef CONFIG_SA1100_ASSABET
   raw_max_x = 944;
   raw_max_y = 944;
   raw_min_x = 70;
   raw_min_y = 70;
   res_x = 320;
   res_y = 240;
#elif defined(CONFIG_SA1100_CERF)
    raw_max_x = 944;
    raw_max_y = 944;
    raw_min_x = 70;
    raw_min_y = 70;
#if defined(CONFIG_CERF_LCD_38_A)
    res_x = 240;
    res_y = 320;
#elif defined(CONFIG_CERF_LCD_57_A)
    res_x = 320;
    res_y = 240;
#elif defined(CONFIG_CERF_LCD_72_A)
    res_x = 640;
    res_y = 480;
#else
#warning "Cannot enable the UCB1200 Touchscreen Driver without selecting a Cerfboard screen orientation first"
#error
#endif

#elif defined(CONFIG_SA1100_FREEBIRD)
   raw_max_x = 925;
   raw_max_y = 875;
   raw_min_x = 85;
   raw_min_y = 60;
   res_x = 240;
   res_y = 320;
#else
   raw_max_x = 885;
   raw_max_y = 885;
   raw_min_x = 70;
   raw_min_y = 70;
   res_x = 640;
   res_y = 480;
#endif

   xyswap = 0;
   head = 0;
   tail = 0;
/* Allen Add */
   cal_ok = 1;
   x_rev = 0;
   y_rev = 0;
/* Allen */

   init_waitqueue_head(&queue);

#ifdef CONFIG_SA1100_ASSABET
   if (machine_is_assabet())
   {
      BCR_set(BCR_CODEC_RST);
      Ser4MCCR0 = MCCR0_MCE | MCCR0_SmpCnt;
      GPDR &= ~GPIO_UCB1300_IRQ;
      set_GPIO_IRQ_edge(GPIO_UCB1300_IRQ, GPIO_BOTH_EDGES);
   }
#endif

#ifdef CONFIG_SA1100_CERF
   if(machine_is_cerf()){
      Ser4MCCR0 = MCCR0_MCE | MCCR0_SmpCnt;
      GPDR &= ~GPIO_UCB1200_IRQ;
      set_GPIO_IRQ_edge(GPIO_UCB1200_IRQ, GPIO_BOTH_EDGES);
   }
#endif

#ifdef CONFIG_SA1100_FREEBIRD
   if (machine_is_freebird())
   {
      BCR_set(BCR_FREEBIRD_CODEC_RST);
      Ser4MCCR0 = MCCR0_MCE | MCCR0_SmpCnt;
      GPDR &= ~GPIO_FREEBIRD_UCB1300;
      set_GPIO_IRQ_edge(GPIO_FREEBIRD_UCB1300, GPIO_BOTH_EDGES);
   }
#endif

#ifdef CONFIG_SA1100_GRAPHICSCLIENT
   if (machine_is_graphicsclient())
   {
      Ser4MCCR0 = MCCR0_MCE;
      set_GPIO_IRQ_edge(GPIO_GPIO22, GPIO_BOTH_EDGES);
   }
#endif

/* Initialize the touchscreen controller */
   codec_write(CODEC_REG_ADC_CTL, 0);
#ifdef CONFIG_SA1100_FREEBIRD
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0 + bt_int_mask);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0 + bt_int_mask);
#else
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
#endif
   codec_write(CODEC_REG_IO_DIRECTION, 0x200);
   ts_clear_codec_int();
   register_chrdev(TS_MAJOR, TS_NAME, &sa1100_ts_fops);

#ifdef CONFIG_SA1100_ASSABET
   if (machine_is_assabet())
      ucb_irq = IRQ_GPIO_UCB1300_IRQ;
#endif

#ifdef CONFIG_SA1100_CERF
   if(machine_is_cerf())
      ucb_irq = IRQ_GPIO_UCB1200_IRQ;
#endif

#ifdef CONFIG_SA1100_FREEBIRD
   if (machine_is_freebird())
      ucb_irq = IRQ_GPIO_FREEBIRD_UCB1300_IRQ;
#endif

#ifdef CONFIG_SA1100_GRAPHICSCLIENT
   if (machine_is_graphicsclient())
      ucb_irq = IRQ_GPIO6;
#endif

   wait_for_action();
   printk("sa1100 touch screen driver initialized\n");
   return 0;
}

void sa1100_ts_cleanup(void)
{
   unsigned int irq;

#ifdef CONFIG_SA1100_ASSABET
   if (machine_is_assabet())
      irq = IRQ_GPIO_UCB1300_IRQ;
#endif

#ifdef CONFIG_SA1100_CERF
   if(machine_is_cerf())
      irq = IRQ_GPIO_UCB1200_IRQ;
#endif

#ifdef CONFIG_SA1100_FREEBIRD
   if (machine_is_freebird())
      irq = IRQ_GPIO_FREEBIRD_UCB1300_IRQ;
#endif

#ifdef CONFIG_SA1100_GRAPHICSCLIENT
   if (machine_is_graphicsclient())
      irq = IRQ_GPIO6;
#endif

   codec_write(CODEC_REG_ADC_CTL, 0);
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   if (in_timehandle)
      del_timer(&timer);
   unregister_chrdev(TS_MAJOR, TS_NAME);

   printk("sa1100 touch screen driver removed\n");
}

#ifdef CONFIG_SA1100_FREEBIRD
// below are routine for button driver
//tony========================================================
static inline void button_clear_codec_int(void)
{
   ushort s;
                                                                                
//   printk("enter button_clear_codec_int =\n");
/* Clear interrupts at codec  */
   if ((s = codec_read(CODEC_REG_INT_STATUS) &  IO0_4_INT  ))
   {
//        printk("clear IO0_INT status\n");
      codec_write(CODEC_REG_INT_STATUS, 0);
      codec_write(CODEC_REG_INT_STATUS, s);
   }
}
                                                                                
                                                                                
static void bt_clear(void)
{
   int i;
                                                                                
   for (i=0; i < BUFSIZE; i++)
   {
       bt_buf[i].keycode=(char)NULL;
       bt_buf[i].down=(int)NULL;
   }
                                                                                
   bt_head = 0;
   bt_tail = 0;
                                                                                
   return 0;
}
                                                                                
static ssize_t sa1100_bt_read(struct file *filp, char *buf, size_t count,loff_t *l)
{
                                                                                
   DECLARE_WAITQUEUE(wait, current);
   int i;
   BT_EVENT t;
   char out_buf[2];
                                                                                
   if (bt_head == bt_tail)
   {
      if (filp->f_flags & O_NONBLOCK)
     return -EAGAIN;
      add_wait_queue(&bt_queue, &wait);
      current->state = TASK_INTERRUPTIBLE;
      while ((bt_head == bt_tail) && !signal_pending(current))
      {
     schedule();
     current->state = TASK_INTERRUPTIBLE;
      }
      current->state = TASK_RUNNING;
      remove_wait_queue(&bt_queue, &wait);
   }
   for (i = count; i >= sizeof(out_buf); i -=sizeof(out_buf), buf+= sizeof(out_buf))
   {
      if (bt_head == bt_tail)
     break;
   //   t = get_key();
                                                                                
      out_buf[0] = bt_buf[bt_tail].keycode;
      out_buf[1] = bt_buf[bt_tail].down;
      bt_tail++;
      copy_to_user(buf, &out_buf, sizeof(out_buf));
   }
   return count - i;
                                                                                
}
                                                                                
static int sa1100_bt_open(struct inode *inode, struct file *filp)
{
   bt_clear();
                                                                                
   MOD_INC_USE_COUNT;
   return 0;
}
                                                                                
static int sa1100_bt_release(struct inode *inode, struct file *filp)
{
   bt_clear();
                                                                                
//   sa1100_ts_fasync(-1, filp, 0);
   MOD_DEC_USE_COUNT;
   return 0;
}
                                                                                
static void sa1110_GPIO0_button_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	sa1110_button_handler(KEY_POWER, (GPLR & GPIO_FREEBIRD_NPOWER_BUTTON));
}

static void sa1110_GPIO1_button_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	sa1110_button_handler(KEY_APP1, (GPLR & GPIO_FREEBIRD_APP1_BUTTON));
}

static void sa1110_GPIO2_button_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	sa1110_button_handler(KEY_APP2, (GPLR & GPIO_FREEBIRD_APP2_BUTTON));
}

static void sa1110_GPIO3_button_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	sa1110_button_handler(KEY_APP3, (GPLR & GPIO_FREEBIRD_APP3_BUTTON));
}

static void sa1110_button_handler(char keycode, int key_state)
{
    ushort down;
                                                                                
    down = ( key_state ) ? 1 : 0;

    bt_buf[bt_head].keycode = keycode;
    bt_buf[bt_head].down = down;
                                                                                
    bt_head++;
    if (bt_head == BUFSIZE)
        bt_head = 0;
                                                                                
    wake_up_interruptible(&bt_queue);
                                                                                
}
                                                                                
//----------------------------
static void button_handler(int irq, void *dev_id, struct pt_regs *regs)
{
    ushort down;
    unsigned short  button,position;
    char    keycode;
	static unsigned long last_button =0; /* record previous button state */

    button = ( codec_read(CODEC_REG_INT_STATUS) & IO0_4_INT );
        // check any button pressed ? if not, return
        if ( button  == 0 )
            return;
    position = ( codec_read(CODEC_REG_IO_DATA) & IO0_4_INT );

	/* when we press the second key without releasing the first key,
	 * UCB1300 sometimes keep the INT status of previously
	 * pressed key. So we have to keep previous key status, and
	 * skip the key when it's a key down and previously pressed key.
	 */
	if ((button & IO0_INT) && 
       (((position & IO0_INT)==0) || ((last_button & IO0_INT)==0))) {
        keycode = KEY_UP;
		down = (position & IO0_INT) ? 1 : 0;
		last_button = (down) ? (last_button|IO0_INT) : (last_button&(~IO0_INT));
	} 

	if ((button & IO1_INT) && 
       (((position & IO1_INT)==0) || ((last_button & IO1_INT)==0))) {
        keycode = KEY_DOWN;
		down = (position & IO1_INT) ? 1 : 0;
		last_button = (down) ? (last_button|IO1_INT) : (last_button&(~IO1_INT));
	}
 
	if ((button & IO2_INT) && 
       (((position & IO2_INT)==0) || ((last_button & IO2_INT)==0))) {
        keycode = KEY_ACTION;
		down = (position & IO2_INT) ? 1 : 0;
		last_button = (down) ? (last_button|IO2_INT) : (last_button&(~IO2_INT));
	}
 
	if ((button & IO3_INT) && 
       (((position & IO3_INT)==0) || ((last_button & IO3_INT)==0))) {
        keycode = KEY_LEFT;
		down = (position & IO3_INT) ? 1 : 0;
		last_button = (down) ? (last_button|IO3_INT) : (last_button&(~IO3_INT));
	}
 
	if ((button & IO4_INT) && 
       (((position & IO4_INT)==0) || ((last_button & IO4_INT)==0))) {
        keycode = KEY_RIGHT;
		down = (position & IO4_INT) ? 1 : 0;
		last_button = (down) ? (last_button|IO4_INT) : (last_button&(~IO4_INT));
	}
	
      button_clear_codec_int();
      bt_buf[bt_head].keycode = keycode;
      bt_buf[bt_head].down = down;
                                                                                
      bt_head++;
      if (bt_head == BUFSIZE)
          bt_head = 0;
                                                                                
      wake_up_interruptible(&bt_queue);
                                                                                
}
                                                                                
static struct file_operations sa1100_bt_fops = {
   read: sa1100_bt_read,
//   ioctl: sa1100_bt_ioctl,
//   fasync: sa1100_bt_fasync,
   open: sa1100_bt_open,
   release: sa1100_bt_release,
};
                                                                                
                                                                                
int sa1100_bt_init(void)
{
   unsigned int irq;
                                                                                
   bt_head = 0;
   bt_tail = 0;
   init_waitqueue_head(&bt_queue);
                                                                                
   GPDR &= ~( 0x00000011);  //set GPIO 0,1 as input
   set_GPIO_IRQ_edge(GPIO_GPIO0, GPIO_BOTH_EDGES);
   set_GPIO_IRQ_edge(GPIO_GPIO1, GPIO_BOTH_EDGES);
   set_GPIO_IRQ_edge(GPIO_GPIO2, GPIO_BOTH_EDGES);
   set_GPIO_IRQ_edge(GPIO_GPIO3, GPIO_BOTH_EDGES);

   button_clear_codec_int();        //tony
   register_chrdev(BN_MAJOR, BN_NAME, &sa1100_bt_fops);
  
// ucb interrupt will be installed in the merged init routine 
                                                                                
   if (request_irq(IRQ_GPIO_FREEBIRD_NPOWER_BUTTON,sa1110_GPIO0_button_handler, 
    SA_INTERRUPT | SA_SAMPLE_RANDOM, "sa1110_power_button", dev_id))
      printk("sa1100_bt_init: button_action fail to register IRQ\n");
                                                                                
   if (request_irq(IRQ_GPIO_FREEBIRD_APP1_BUTTON, sa1110_GPIO1_button_handler,
    SA_INTERRUPT | SA_SAMPLE_RANDOM, "sa1110_app1_button", dev_id))
      printk("sa1100_bt_init: button_action fail to register IRQ\n");
                                                                                
   if (request_irq(IRQ_GPIO_FREEBIRD_APP2_BUTTON, sa1110_GPIO2_button_handler,
    SA_INTERRUPT | SA_SAMPLE_RANDOM, "sa1110_app2_button", dev_id))
      printk("sa1100_bt_init: button_action fail to register IRQ\n");

   if (request_irq(IRQ_GPIO_FREEBIRD_APP3_BUTTON, sa1110_GPIO3_button_handler, 
    SA_INTERRUPT | SA_SAMPLE_RANDOM, "sa1110_app3_button", dev_id))
      printk("sa1100_bt_init: button_action fail to register IRQ\n");

   printk("sa1100 button initialized\n");
   return 0;
}
                                                                                
void sa1100_bt_cleanup(void)
{
   unsigned int irq;
                                                                                
   codec_write(CODEC_REG_RISE_INT_ENABLE, 0);
   codec_write(CODEC_REG_FALL_INT_ENABLE, 0);
   free_irq(IRQ_GPIO0, sa1110_button_handler);
   free_irq(IRQ_GPIO1, sa1110_button_handler);
   unregister_chrdev(BN_MAJOR, BN_NAME);
                                                                                
   printk("sa1100 button driver removed\n");
}

#endif

// below are merged routines for touch screen and button
static void ucb_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned short int_status;

	int_status = codec_read(CODEC_REG_INT_STATUS);
#ifdef CONFIG_SA1100_FREEBIRD
	if (int_status & IO0_4_INT)
		button_handler(irq, dev_id, regs);
#endif
	if (int_status & (TSPX_INT | ADC_INT))
		sa1100_ts_interrupt(irq, dev_id, regs);
}

int __init sa1100_ucb_init(void)
{
	sa1100_ts_init();
#ifdef CONFIG_SA1100_FREEBIRD
	sa1100_bt_init();
#endif
   if (request_irq(ucb_irq, ucb_int_handler, SA_INTERRUPT | SA_SHIRQ, TS_NAME, dev_id))
      printk("sa1100_ts_init: failed to register IRQ\n");
   return 0;
}

void __exit sa1100_ucb_cleanup(void)
{
	sa1100_ts_cleanup();
#ifdef CONFIG_SA1100_FREEBIRD
	sa1100_bt_cleanup();
#endif
    free_irq(ucb_irq, dev_id);
}

module_init(sa1100_ucb_init);
module_exit(sa1100_ucb_cleanup);
